import numpy as np
import random
from fractions import Fraction
from poisson_approval.utils.DictPrintingInOrder import DictPrintingInOrder


def initialize_random_seeds(n=0):
    """Initialize the random seeds.

    Parameters
    ----------
    n : int
        The desired random seed. Default: 0.
    """
    random.seed(n)
    np.random.seed(n)


def rand_simplex(d=6):
    """Draw a random point in the simplex.

    Parameters
    ----------
    d : int
        Number of coordinates. In other words, we consider the simplex of dimension `d - 1`.

    Returns
    -------
    numpy.ndarray
        A numpy array of length `d`, whose sum is 1.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> rand_simplex(d=6)  # doctest: +SKIP
        array([0.4236548 , 0.12122838, 0.00393032, 0.05394987, 0.11242599, 0.28481063])
    """
    x = np.sort(np.random.rand(d - 1))
    return np.concatenate((x, [1])) - np.concatenate(([0], x))


def rand_integers_fixed_sum(n_integers, fixed_sum):
    """Generate integers with a given sum (uniformly).

    Parameters
    ----------
    n_integers : int
        The desired number of integers.
    fixed_sum : int
        The fixed sum.

    Returns
    -------
    numpy.ndarray
        A numpy array of `d` integers, whose sum is `fixed_sum`, and drawn uniformly.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> rand_integers_fixed_sum(n_integers=6, fixed_sum=100)
        array([ 2, 23, 34,  0, 22, 19])
    """
    n_separators = n_integers - 1
    separators = np.concatenate((
        [-1],
        np.sort(np.random.choice(fixed_sum + n_separators, n_separators, replace=False)),
        [fixed_sum + n_separators]))
    return np.array([up - down - 1 for down, up in zip(separators[:-1], separators[1:])])


def rand_simplex_grid(d, denominator):
    """Draw a random point in the simplex, with rational coordinates of a given denominator

    Parameters
    ----------
    d : int
        Number of coordinates. In other words, we consider the simplex of dimension `d - 1`.
    denominator : int
        The coordinates will be fractions with this denominator.

    Returns
    -------
    numpy.ndarray
        A numpy array of length `d`, whose coordinates are fractions of the given denominator, and whose sum is 1.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> rand_simplex_grid(d=3, denominator=100)
        array([Fraction(13, 50), Fraction(13, 20), Fraction(9, 100)], dtype=object)
    """
    return np.array([
        Fraction(int(n), denominator)
        for n in rand_integers_fixed_sum(n_integers=d, fixed_sum=denominator)])


def probability(generator, n_samples, test, conditional_on=None):
    """Probability that a random `something` meets some given test.

    Parameters
    ----------
    generator : callable or tuple
        This can be:

        * Either a callable that takes no input and that outputs a (random) `something`,
        * Or a tuple of such generators (cf. examples below).
    n_samples : int
        Number of samples.
    test : callable
        A function that take as input(s) the output(s) of the generator(s) and that returns a Boolean.
    conditional_on : callable
        A function that take as input(s) the output(s) of the generator(s) and that returns a Boolean.
        Default: always True.

    Returns
    -------
    float
        The probability that the output(s) generated by `generator` meet(s) `test`, conditional on the fact
        that it meets `conditional_on`, based on a Monte-Carlo estimation of `n_samples` trials.

    Examples
    --------
    In this basic example with one generator, we estimate the probability that a random float between 0 and 1 is greater
    than .5, conditionally on being greater than .25:

        >>> import random
        >>> initialize_random_seeds()
        >>> def generator():
        ...     return random.random()
        >>> probability(generator, n_samples=1000, test=lambda x: x > .5, conditional_on=lambda x: x > .25)
        0.661

    In this example with a tuple of generators, we estimate the probability that a random 2*2 matrix and a random vector
    of size 2, both with integer coefficients between -10 included and 11 excluded, have a dot product that is null,
    conditionally on not being null themselves:

        >>> import numpy as np
        >>> initialize_random_seeds()
        >>> def generator_matrix():
        ...     return np.random.randint(-10, 11, (2, 2))
        >>> def generator_vector():
        ...     return np.random.randint(-10, 11, 2)
        >>> def test_dot_zero(matrix, vector):
        ...     return np.all(np.dot(matrix, vector) == 0)
        >>> def test_non_trivial(matrix, vector):
        ...     return not np.all(matrix == 0) and not np.all(vector == 0)
        >>> probability(generator=(generator_matrix, generator_vector), n_samples=10000,
        ...             test=test_dot_zero, conditional_on=test_non_trivial)
        0.0003
    """
    if not isinstance(generator, tuple):
        generator = (generator, )
    success = 0
    i_samples = 0
    while i_samples < n_samples:
        somethings = [g() for g in generator]
        if conditional_on is None or conditional_on(*somethings):
            i_samples += 1
            if test(*somethings):
                success += 1
    return success / n_samples


def image_distribution(generator, n_samples, f, conditional_on=None):
    """Distribution of ``f(something)`` for a random `something`.

    Parameters
    ----------
    generator : callable or tuple
        This can be:

        * Either a callable that takes no input and that outputs a (random) `something`,
        * Or a tuple of such generators (cf. examples below).
    n_samples : int
        Number of samples.
    f : callable
        A function that take as input(s) the output(s) of the generator(s)..
    conditional_on : callable
        A function that take as input(s) the output(s) of the generator(s) and that returns a Boolean.
        Default: always True.

    Returns
    -------
    DictPrintingInOrder
        Keys: the obtained outputs for `f`. Values: the probability that `f(something)` has this output when `something`
        is generated by `generator`, conditional on the fact that it meets `conditional_on`,
        based on a Monte-Carlo estimation of `n_samples` trials.

    Examples
    --------
    In this basic example with one generator, we compute the distribution of `n` modulo 10, when `n` is drawn uniformly
    at random between 0 included and 100 excluded, conditionally on the fact that `n` is even:

        >>> import numpy as np
        >>> initialize_random_seeds()
        >>> def generator_integer():
        ...     return np.random.randint(0, 100)
        >>> def modulo_10(n):
        ...     return n % 10
        >>> def is_even(n):
        ...     return n % 2 == 0
        >>> image_distribution(generator=generator_integer, n_samples=100, f=modulo_10, conditional_on=is_even)
        {0: 0.21, 2: 0.21, 4: 0.2, 6: 0.18, 8: 0.2}

    In this example with a tuple of generator, we compute the distribution of `a mod b`, when `a` is drawn uniformly
    at random between 0 included and 100 excluded, and `b` is drawn uniformly at random between 1 included and 11
    excluded:

        >>> import numpy as np
        >>> initialize_random_seeds()
        >>> def generator_integer():
        ...     return np.random.randint(0, 100)
        >>> def generator_divider():
        ...     return np.random.randint(1, 11)
        >>> def modulo(a, b):
        ...     return a % b
        >>> image_distribution(generator=(generator_integer, generator_divider),
        ...                    n_samples=100, f=modulo)
        {0: 0.31, 1: 0.16, 2: 0.18, 3: 0.12, 4: 0.07, 5: 0.04, 6: 0.02, 7: 0.08, 9: 0.02}
    """
    if not isinstance(generator, tuple):
        generator = (generator, )
    d_result_occurrences = dict()
    i_samples = 0
    while i_samples < n_samples:
        somethings = [g() for g in generator]
        if conditional_on is None or conditional_on(*somethings):
            i_samples += 1
            result = f(*somethings)
            d_result_occurrences[result] = d_result_occurrences.get(result, 0) + 1
    return DictPrintingInOrder({result: occurrences / n_samples
                                for result, occurrences in d_result_occurrences.items()})


def _false_for_fraction(f):
    """Decorator to return False when the input is a Fraction (cf. usages below)."""
    def _f(x):
        if isinstance(x, Fraction):
            return False
        return f(x)
    _f.__doc__ = f.__doc__
    return _f


@_false_for_fraction
def isnan(x):
    """Is nan.

    Parameters
    ----------
    x : Number

    Returns
    -------
    bool
        True if `x` is nan.

    Notes
    -----
    This extends the usual numpy function ``isnan`` to fractions.

    Examples
    --------
        >>> from fractions import Fraction
        >>> isnan(Fraction(1, 10))
        False
        >>> isnan(1)
        False
        >>> isnan(np.nan)
        True
    """
    return np.isnan(x)


@_false_for_fraction
def isposinf(x):
    """Is positive infinity.

    Parameters
    ----------
    x : Number

    Returns
    -------
    bool
        True if `x` is positive infinity.

    Notes
    -----
    This extends the usual numpy function ``isposinf`` to fractions.

    Examples
    --------
        >>> from fractions import Fraction
        >>> isposinf(Fraction(1, 10))
        False
        >>> isposinf(1)
        False
        >>> isposinf(np.inf)
        True
    """
    return np.isposinf(x)


@_false_for_fraction
def isneginf(x):
    """Is negative infinity.

    Parameters
    ----------
    x : Number

    Returns
    -------
    bool
        True if `x` is negative infinity.

    Notes
    -----
    This extends the usual numpy function ``isneginf`` to fractions.

    Examples
    --------
        >>> from fractions import Fraction
        >>> isposinf(Fraction(1, 10))
        False
        >>> isneginf(1)
        False
        >>> isneginf(- np.inf)
        True
    """
    return np.isneginf(x)


def sort_ballot(ballot):
    """Put a ballot in alphabetical order.

    Parameters
    ----------
    ballot : str
        A ballot, e.g. ``'a'``, ``'ab'``, ``'ba'``, etc.

    Returns
    -------
    str
        The same ballot in alphabetical order.

    Examples
    --------
        >>> sort_ballot('a')
        'a'
        >>> sort_ballot('ab')
        'ab'
        >>> sort_ballot('ba')
        'ab'
    """
    return ''.join(sorted(ballot))


def ballot_one(ranking):
    """Ballot for the voter's preferred candidate.

    Parameters
    ----------
    ranking : str
        A ranking.

    Returns
    -------
    str
        The first candidate.

    Examples
    --------
        >>> ballot_one('abc')
        'a'
    """
    return ranking[0]


def ballot_one_two(ranking):
    """Ballot for the voter's two preferred candidates.

    Parameters
    ----------
    ranking : str
        A ranking.

    Returns
    -------
    str
        The ballot with the two first candidates.

    Examples
    --------
        >>> ballot_one_two('abc')
        'ab'
        >>> ballot_one_two('bac')
        'ab'
    """
    return min(ranking[:2], ranking[1::-1])


def give_figure(n, singular, plural=None):
    """Combine a number with a unit, whose word can be singular or plural.

    Parameters
    ----------
    n : int
        The number.
    singular : str
        The singular word.
    plural : str
        The plural word. Default: the singular with a final 's'.

    Returns
    -------
    str
        The number and the word.

    Examples
    --------
        >>> give_figure(1, 'apple')
        '1 apple'
        >>> give_figure(2, 'apple')
        '2 apples'
        >>> give_figure(1, 'equilibrium', 'equilibria')
        '1 equilibrium'
        >>> give_figure(2, 'equilibrium', 'equilibria')
        '2 equilibria'
    """
    if n <= 1:
        return str(n) + ' ' + singular
    else:
        if plural is None:
            return str(n) + ' ' + singular + 's'
        else:
            return str(n) + ' ' + plural


def barycenter(a, b, ratio_b):
    """Barycenter.

    Parameters
    ----------
    a : Number
    b : Number
    ratio_b : the ratio of `b` in the result

    Returns
    -------
    Number
        The result of ``(1 - ratio_b) * a + ratio_b * b``. The added value of this function is to preserve the type
        of `a` (resp. `b`) when `ratio_b` is 0 (resp. 1).

    Examples
    --------
    In this first example, `barycenter` preserves the type Fraction, whereas a naive computation returns a float:

        >>> a, b = Fraction(1, 10), 0.7
        >>> ratio_b = 0
        >>> barycenter(a, b, ratio_b)
        Fraction(1, 10)
        >>> (1 - ratio_b) * a + ratio_b * b
        0.1

    The second example is symmetric of the first one, in the sense that it preserves the type of `b`:

        >>> a, b = 0.7, 42
        >>> ratio_b = 1
        >>> barycenter(a, b, ratio_b)
        42
        >>> (1 - ratio_b) * a + ratio_b * b
        42.0
    """
    if ratio_b == 0:
        return a
    if ratio_b == 1:
        return b
    return (1 - ratio_b) * a + ratio_b * b



from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event
from poisson_approval.events.EventPivotWeak import EventPivotWeak


class EventPivotTij(Event):
    """A `personalized pivot` of type ``Tij``.

    Notes
    -----
    We consider the personalized pivot for the two most-liked candidates of a voter ``xyz``, i.e. situations
    where, if the voter add a reasonable ballot (``x`` or ``xy``), it become a strict pivot for ``xy``. In other words,
    situations where ``S_x + 1 = S_y > S_z`` or ``S_x + 1 = S_y + 1 > S_z``.

    For parameters and attributes, cf. :class:`Event`.

    Examples
    --------
        >>> from fractions import Fraction
        >>> EventPivotTij(candidate_x='a', candidate_y='b', candidate_z='c',
        ...               tau_a=Fraction(1, 10), tau_ab=Fraction(6, 10), tau_c=Fraction(3, 10))
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    """

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        pivot_weak = EventPivotWeak(candidate_x='x', candidate_y='y', candidate_z='z',
                                    tau_x=tau_x, tau_y=tau_y, tau_z=tau_z, tau_xy=tau_xy, tau_xz=tau_xz, tau_yz=tau_yz)
        self._phi_x = pivot_weak.phi['x']
        self._phi_y = pivot_weak.phi['y']
        self._phi_z = pivot_weak.phi['z']
        self._phi_xy = pivot_weak.phi['xy']
        self._phi_xz = pivot_weak.phi['xz']
        self._phi_yz = pivot_weak.phi['yz']
        if tau_x == 0 and tau_xy == 0:
            # Flower diagram: holes `at the bottom`, i.e. around ``xy`` (``x`` side)
            self.asymptotic = (Asymptotic.poisson_value(tau_yz, 0)
                               * (Asymptotic.poisson_one_more(tau_y, tau_xz) + Asymptotic.poisson_eq(tau_y, tau_xz))
                               * Asymptotic.poisson_value(tau_z, 0))
        elif tau_y == 0 and tau_xy == 0:
            # Flower diagram: holes `at the bottom`, i.e. around ``xy`` (``y`` side)
            self.asymptotic = (Asymptotic.poisson_eq(tau_x, tau_yz) * Asymptotic.poisson_value(tau_xz, 0)
                               * Asymptotic.poisson_value(tau_z, 0))
        elif tau_y == 0 and tau_yz == 0:
            # Flower diagram: consecutive holes on the ``y`` side
            self.asymptotic = (Asymptotic.poisson_value(tau_x, 0) * Asymptotic.poisson_value(tau_xz, 0)
                               * Asymptotic.poisson_ge(tau_xy, tau_z))
        elif tau_x == 0 and tau_xz == 0:
            # Flower diagram: consecutive holes on the ``x`` side
            self.asymptotic = (
                Asymptotic.poisson_value(tau_yz, 0)
                * (Asymptotic.poisson_value(tau_y, 0) + Asymptotic.poisson_value(tau_y, 1))
                * Asymptotic.poisson_ge(tau_xy, tau_z)
            ) + (
                Asymptotic.poisson_value(tau_yz, 1) * Asymptotic.poisson_value(tau_y, 0)
                * Asymptotic.poisson_gt(tau_xy, tau_z)
            )
        elif tau_z == 0 and (tau_xz == 0 or tau_yz == 0):
            # Flower diagram: holes `at the top`, i.e. around ``z``
            self.asymptotic = (Asymptotic.poisson_one_more(tau_y + tau_yz, tau_x + tau_xz)
                               + Asymptotic.poisson_eq(tau_y + tau_yz, tau_x + tau_xz))
        else:
            _phi_xz_tilde = self._phi_xz if tau_xz != 0 else self._phi_x * self._phi_z
            self.asymptotic = pivot_weak.asymptotic * (1 + _phi_xz_tilde)



from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import initialize_random_seeds, rand_simplex_grid
from poisson_approval.tau_vector.TauVector import TauVector


class GeneratorTauVectorGridUniform:
    """A generator of tau-vectors (:class:`TauVector`), uniform on a grid.

    Parameters
    ----------
    denominator : int
        The coefficients of the tau-vector will be fractions with this denominator.

    Notes
    -----
    The tau-vector is drawn uniformly on the points of the simplex whose coordinates are fractions with the given
    denominator.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> generator = GeneratorTauVectorGridUniform(denominator=100)
        >>> tau = generator()
        >>> print(tau)
        <a: 1/50, ac: 11/50, b: 23/100, bc: 19/100, c: 17/50> ==> c
    """

    def __init__(self, denominator):
        self.denominator = denominator

    def __call__(self):
        """
        Returns
        -------
        TauVector
            A tau-vector.
        """
        x = rand_simplex_grid(d=6, denominator=self.denominator)
        return TauVector({ballot: x[i] for i, ballot in enumerate(BALLOTS_WITHOUT_INVERSIONS)})



from math import isclose
from poisson_approval.constants.constants import *
from poisson_approval.strategies.StrategyTwelve import StrategyTwelve
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros


# noinspection PyUnresolvedReferences
class StrategyThreshold(StrategyTwelve):
    """A strategy profile for a cardinal profile.

    Parameters
    ----------
    d_ranking_threshold : dict
        Keys are rankings and values are utility thresholds. E.g. ``'abc': 0.4`` means that a voter ``abc`` vote for
        ``ab`` if her utility for ``b`` is strictly greater than 0.4, and for ``a`` otherwise. A threshold can be
        ``None``, meaning that the behavior of these voters is not specified in the strategy.
    profile : Profile, optional
        The "context" in which the strategy is used.

    Examples
    --------
        >>> strategy = StrategyThreshold({'abc': 0.4, 'bac': 0.51, 'cab': 1})
        >>> strategy
        StrategyThreshold({'abc': 0.4, 'bac': 0.51, 'cab': 1})
        >>> print(strategy)
        <abc: utility-dependent (0.4), bac: utility-dependent (0.51), cab: c>
        >>> strategy.abc
        'utility-dependent'
        >>> strategy.a_bc
        'a'
        >>> strategy.ab_c
        'ab'
        >>> strategy.d_ranking_threshold['abc']
        0.4
    """

    def __init__(self, d_ranking_threshold, profile=None):
        # Populate the dictionary of thresholds
        self.d_ranking_threshold = DictPrintingInOrderIgnoringZeros(d_ranking_threshold)
        for ranking in RANKINGS:
            if ranking not in self.d_ranking_threshold:
                self.d_ranking_threshold[ranking] = None
        # Prepare the dictionary of ballots
        d_ranking_ballot = DictPrintingInOrderIgnoringZeros()
        for ranking, threshold in self.d_ranking_threshold.items():
            if threshold is None:
                d_ranking_ballot[ranking] = ''
            elif threshold == 1:
                d_ranking_ballot[ranking] = ranking[0]
            elif threshold == 0:
                d_ranking_ballot[ranking] = ranking[:2]
            else:
                d_ranking_ballot[ranking] = UTILITY_DEPENDENT
        # Call parent class
        super().__init__(d_ranking_ballot=d_ranking_ballot, profile=profile)

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : object

        Returns
        -------
        bool
            True if this strategy is equal to `other`.

        Examples
        --------
            >>> strategy = StrategyThreshold({'abc': 0.4, 'bac': 0.51, 'cab': 1})
            >>> strategy == StrategyThreshold({'abc': 0.4, 'bac': 0.51, 'cab': 1})
            True
        """
        return isinstance(other, StrategyThreshold) and self.d_ranking_threshold == other.d_ranking_threshold

    def isclose(self, other, *args, **kwargs):
        """Test near-equality.

        Parameters
        ----------
        other : Object
        *args
            Cf. :func:`math.isclose`.
        **kwargs
            Cf. :func:`math.isclose`.

        Returns
        -------
        isclose : bool
            True if this strategy is approximately equal to `other`.

        Examples
        --------
            >>> strategy = StrategyThreshold({'abc': 0.4, 'bac': 0.51, 'cab': 1})
            >>> strategy.isclose(StrategyThreshold({'abc': 0.4, 'bac': 0.51, 'cab': 0.999999999999999999999999}))
            True
        """
        return isinstance(other, StrategyThreshold) and all([
            (threshold is None and other.d_ranking_threshold[ranking] is None)
            or isclose(threshold, other.d_ranking_threshold[ranking], *args, **kwargs)
            for ranking, threshold in self.d_ranking_threshold.items()
        ])

    def __repr__(self):
        return 'StrategyThreshold(%r)' % self.d_ranking_threshold

    def __str__(self):
        result = '<' + ', '.join([
            '%s: %s' % (ranking, self.d_ranking_ballot[ranking])
            + (' (%s)' % self.d_ranking_threshold[ranking]
               if self.d_ranking_ballot[ranking] == UTILITY_DEPENDENT else '')
            for ranking in sorted(self.d_ranking_ballot) if self.d_ranking_ballot[ranking]
        ]) + '>'
        if self.profile is not None:
            result += ' ==> ' + str(self.winners)
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')



import numpy as np


class DictPrintingInOrderIgnoringZeros(dict):
    """A dictionary that prints in the order of the keys, ignoring values such as 0, None...

    Notes
    -----
    All values whose boolean conversion is False are ignored when converting to string: 0, None, [], all empty
    containers, etc. Empty numpy arrays are also ignored.

    Examples
    --------
        >>> d = DictPrintingInOrderIgnoringZeros({'b': 'x', 'a': 'y', 'c': 'z', 'd': 0, 'e': None, 'f': []})
        >>> print(d)
        {a: y, b: x, c: z}
        >>> print(repr(d))
        {'a': 'y', 'b': 'x', 'c': 'z'}
    """

    def __repr__(self):
        return "{" + ", ".join([
            "%r: %r" % (key, self[key]) for key in sorted(self)
            if (isinstance(self[key], np.ndarray) and len(self[key]) > 0) or (
                not isinstance(self[key], np.ndarray) and self[key])
        ]) + "}"

    def __str__(self):
        return "{" + ", ".join([
            "%s: %s" % (key, self[key]) for key in sorted(self)
            if (isinstance(self[key], np.ndarray) and len(self[key]) > 0) or (
                not isinstance(self[key], np.ndarray) and self[key])
        ]) + "}"

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')



import numpy as np
import pytest
from poisson_approval import Asymptotic


def test():
    assert np.isnan(Asymptotic(mu=np.nan, nu=51, xi=69).limit)
    assert np.isposinf(Asymptotic(mu=0, nu=51, xi=69).limit)

    asymptotic = Asymptotic(mu=np.nan, nu=51, xi=69) + Asymptotic(mu=42, nu=51, xi=69)
    assert np.isnan(asymptotic.mu)
    assert np.isnan(asymptotic.nu)
    assert np.isnan(asymptotic.xi)

    assert not Asymptotic(mu=42, nu=51, xi=69).isclose(144)
    with pytest.raises(ValueError):
        Asymptotic(mu=np.nan, nu=51, xi=69).isclose(Asymptotic(mu=42, nu=51, xi=69))



from poisson_approval.utils.Util import isnan


class Event:
    """An event with all its attributes: magnitudes, offsets, asymptotic if possible (abstract class).

    Parameters
    ----------
    candidate_x : str
        A candidate (e.g. ``'a'``).
    candidate_y : str
        A candidate (e.g. ``'b'``).
    candidate_z : str
        A candidate (e.g. ``'c'``).
    kwargs
        Use ``tau_..`` for the values of the tau-vector (probability for each kind of ballot). For example, ``tau_a``,
        ``tau_ab``, etc.

    Attributes
    ----------
    asymptotic : Asymptotic
        The asymptotic development of the probability of the event when `n` tends to infinity.
    mu : Number, ``np.nan`` or ``np.inf``
        Shortcut for ``asymptotic.mu``.
    nu : Number, ``np.nan`` or ``np.inf``
        Shortcut for ``asymptotic.nu``.
    xi : Number, ``np.nan`` or ``np.inf``
        Shortcut for ``asymptotic.xi``.
    phi_ab : Number or ``np.nan``
        The offset for this kind of ballot. An offset is ``np.nan`` if it is not defined.
        Other offsets are denoted ``phi_a``, etc.

    Notes
    -----
    The permutation of ``candidate_x``, ``candidate_y`` and ``candidate_z`` is generally important in the subclasses.
    For example, in :class:`EventDuo`, we consider a tie between candidates ``x`` and ``y`` (hence in that case, if is
    the same if ``x`` and ``y`` are exchanged, but not if ``x`` and ``z`` are exchanged for example).

    Examples
    --------
    Cf. :class:`EventPivotWeak`.
    """

    def __init__(self, candidate_x, candidate_y, candidate_z, **kwargs):
        """
        Examples
        --------
            >>> event = Event(candidate_x='a', candidate_y='b', candidate_z='c', non_existing_argument=42)
            Traceback (most recent call last):
            ValueError: ('Unknown arguments: ', ['non_existing_argument'])
        """
        # -------------
        # Preliminaries
        # -------------
        # Create labels
        self._label_x, self._label_y, self._label_z = candidate_x, candidate_y, candidate_z
        self._label_xy = ''.join(sorted([self._label_x, self._label_y]))
        self._label_xz = ''.join(sorted([self._label_x, self._label_z]))
        self._label_yz = ''.join(sorted([self._label_y, self._label_z]))
        self._label_xyd = self._label_xy[1] + self._label_xy[0]
        self._label_xzd = self._label_xz[1] + self._label_xz[0]
        self._label_yzd = self._label_yz[1] + self._label_yz[0]
        self._labels_std_one = {self._label_x: 'x', self._label_y: 'y', self._label_z: 'z'}
        self._labels_std_two = {self._label_xy: 'xy', self._label_xz: 'xz', self._label_yz: 'yz'}
        self._labels_std_two_down = {self._label_xyd: 'xy', self._label_xzd: 'xz', self._label_yzd: 'yz'}
        self._labels_std = self._labels_std_one.copy()
        self._labels_std.update(self._labels_std_two)
        self._labels_std.update(self._labels_std_two_down)
        # Declare variables to tranquilize PyCharm's syntax checker
        self._tau_x, self._tau_y, self._tau_z = 0, 0, 0
        self._tau_xy, self._tau_xz, self._tau_yz = 0, 0, 0
        # Initialize variables such as self.tau_a and self._tau_x
        for label, label_std in self._labels_std.items():
            # Ex: label = 'ab', label_std = 'xy'
            # Update variable such as self._tau_xy if provided.
            if 'tau_' + label in kwargs.keys():
                setattr(self, '_tau_' + label_std, kwargs.pop('tau_' + label))
            # Anyway, define variable such as tau_ab (0 if not provided).
            setattr(self, 'tau_' + label, getattr(self, '_tau_' + label_std))
        # Check that no unused argument remains
        if kwargs:
            raise ValueError('Unknown arguments: ', list(kwargs.keys()))
        # Declare the computed variables
        self._phi_x, self._phi_y, self._phi_z = None, None, None
        self._phi_xy, self._phi_xz, self._phi_yz = None, None, None
        self.asymptotic = None
        # ------------------------------------------------
        # Magnitudes, offsets and (if possible) asymptotic
        # ------------------------------------------------
        self._compute(tau_x=self._tau_x, tau_y=self._tau_y, tau_z=self._tau_z,
                      tau_xy=self._tau_xy, tau_xz=self._tau_xz, tau_yz=self._tau_yz)
        self.mu = self.asymptotic.mu
        self.nu = self.asymptotic.nu
        self.xi = self.asymptotic.xi
        # -----------------------------------------------------
        # Create the variables like self.phi_ab, self.phi['ab']
        # -----------------------------------------------------
        self.phi = dict()
        for label, label_std in self._labels_std.items():
            # Ex: label = 'ab', label_std = 'xy'
            setattr(self, 'phi_' + label, getattr(self, '_phi_' + label_std))
            self.phi[label] = getattr(self, '_phi_' + label_std)

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        """
        Compute the magnitude, offsets and (if possible) asymptotic.

        Should update:

            * ``self.asymptotic`` (with, at least, the magnitude).
            * ``self._phi_x``, `self._phi_y``, `self._phi_z``, `self._phi_xy``, `self._phi_xz``, `self._phi_yz``,
        """
        raise NotImplementedError

    def __repr__(self):
        s = 'asymptotic = %s' % self.asymptotic
        lab_sorted = sorted(self._labels_std_one.keys())
        lab_sorted.extend(sorted(self._labels_std_two.keys()))
        for label in lab_sorted:
            val = getattr(self, 'phi_' + label)
            if not isnan(val):
                s += ', phi_' + label + ' = {:.6g}'.format(float(val))
        return '<%s>' % s

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')



from fractions import Fraction
from poisson_approval import TauVector, BestResponse


def test():
    # Test if normalization works as expected
    tau = TauVector({'a': 1, 'b': 1})
    assert tau.a == 0.5

    # Test a case where the justification is easy vs difficult pivot (or the contrary)
    tau = TauVector({'a': Fraction(1, 3), 'ab': Fraction(1, 3), 'b': Fraction(1, 6), 'c': Fraction(1, 6)})
    assert tau.d_ranking_best_response['acb'].ballot == 'a'
    assert tau.d_ranking_best_response['bca'].ballot == 'bc'



from math import isclose
from poisson_approval.utils.Util import isnan, ballot_one, ballot_one_two
from poisson_approval.constants.constants import *


class BestResponse:
    """Best response for a given ordinal type of voter.

    Parameters
    ----------
    ranking : str
        Voter's ranking, e.g. ``'abc'``.
    pivot_tij : EventPivotTij
        `Personalized pivot` between her first and second candidate.
    pivot_tjk : EventPivotTjk
        `Personalized pivot` between her second and third candidate.
    trio_1t : EventTrio1t
        First `personalized trio`.
    trio_2t : EventTrio2t
        Second `personalized trio`.
    trio : EventTrio
        3-candidate tie.

    Attributes
    ----------
    ballot : str
        This can be a valid ballot (e.g. ``'a'`` or ``'ab'`` if `ranking` is ``'abc'``) or ``'utility-dependent'``.
        Historically, it could also be ``'inconclusive'``, but this value is not used anymore in the current version
        of the algorithm.
    threshold_utility : Number
        The threshold value of the utility for the second candidate (where the optimal ballot changes). If 1, then
        always vote for the first candidate. If 0, then always vote for the two most-liked candidates.
    justification : str
        How the program made its decision. Nowadays, possible values are ``'Asymptotic method'``,
        ``'Simplified asymptotic method'``, ``'Easy vs difficult pivot'``, ``'Difficult vs easy pivot'``,
        ``'Offset method'``.

    Examples
    --------
        >>> from poisson_approval import EventPivotTij, EventPivotTjk, EventTrio1t, EventTrio2t, EventTrio
        >>> tau = {'tau_a': 9/15, 'tau_b': 4/15, 'tau_ac': 1/15, 'tau_bc': 1/15}
        >>> pivot_tij = EventPivotTij(candidate_x='a', candidate_y='b', candidate_z='c', **tau)
        >>> pivot_tjk = EventPivotTjk(candidate_x='c', candidate_y='b', candidate_z='a', **tau)
        >>> trio_1t = EventTrio1t(candidate_x='a', candidate_y='b', candidate_z='c', **tau)
        >>> trio_2t = EventTrio2t(candidate_x='a', candidate_y='b', candidate_z='c', **tau)
        >>> trio = EventTrio(candidate_x='a', candidate_y='b', candidate_z='c', **tau)
        >>> best_response = BestResponse('abc', pivot_tij, pivot_tjk, trio_1t, trio_2t, trio)
        >>> best_response.ballot
        'a'
        >>> best_response.threshold_utility
        1.0
    """

    ASYMPTOTIC = 'Asymptotic method'
    ASYMPTOTIC_SIMPLIFIED = 'Simplified asymptotic method'
    EASY_VS_DIFFICULT = 'Easy vs difficult pivot'
    DIFFICULT_VS_EASY = 'Difficult vs easy pivot'
    OFFSET_METHOD = 'Offset method'

    def __init__(self, ranking, pivot_tij, pivot_tjk, trio_1t, trio_2t, trio):
        # -------------
        # Preliminaries
        # -------------
        self.ranking = ranking
        self.pivot_tij = pivot_tij
        self.pivot_tjk = pivot_tjk
        self.trio_1t = trio_1t
        self.trio_2t = trio_2t
        self.trio = trio
        # -----------
        # Actual work
        # -----------
        self.threshold_utility = ((
            pivot_tij.asymptotic * (1 / 2)
            + trio_1t.asymptotic * (1 / 3) + trio_2t.asymptotic * (1 / 6)
        ) / (
            pivot_tij.asymptotic * (1 / 2) + pivot_tjk.asymptotic * (1 / 2)
            + trio_1t.asymptotic * (2 / 3) + trio_2t.asymptotic * (1 / 3)
        )).limit
        self.justification = self.ASYMPTOTIC
        if isnan(self.threshold_utility):
            phi_i = getattr(trio, 'phi_' + ranking[0])
            if isnan(phi_i):
                phi_i = (getattr(trio, 'phi_' + ranking[0] + ranking[1])
                         * getattr(trio, 'phi_' + ranking[0] + ranking[2]))
            phi_j = getattr(trio, 'phi_' + ranking[1])
            if isnan(phi_j):
                phi_j = (getattr(trio, 'phi_' + ranking[1] + ranking[0])
                         * getattr(trio, 'phi_' + ranking[1] + ranking[2]))
            phi_k = getattr(trio, 'phi_' + ranking[2])
            if isnan(phi_k):
                phi_k = (getattr(trio, 'phi_' + ranking[2] + ranking[0])
                         * getattr(trio, 'phi_' + ranking[2] + ranking[1]))
            phi_ij = getattr(trio, 'phi_' + ranking[0] + ranking[1])
            if isnan(phi_ij):
                phi_ij = getattr(trio, 'phi_' + ranking[0]) * getattr(trio, 'phi_' + ranking[1])
            phi_ik = getattr(trio, 'phi_' + ranking[0] + ranking[2])
            if isnan(phi_ik):
                phi_ik = getattr(trio, 'phi_' + ranking[0]) * getattr(trio, 'phi_' + ranking[2])
            phi_i_ge_1 = (isclose(phi_i, 1) or phi_i >= 1)
            phi_k_ge_1 = (isclose(phi_k, 1) or phi_k >= 1)
            if phi_i_ge_1 and phi_k_ge_1:
                # Both pivots are easy: we can forget the trios
                self.threshold_utility = ((
                    pivot_tij.asymptotic * (1 / 2)
                ) / (
                    pivot_tij.asymptotic * (1 / 2) + pivot_tjk.asymptotic * (1 / 2)
                )).limit
                self.justification = self.ASYMPTOTIC_SIMPLIFIED
            elif phi_i_ge_1:
                # Pivot jk is difficult, pivot ij is easy
                self.threshold_utility = 0
                self.justification = self.EASY_VS_DIFFICULT
            elif phi_k_ge_1:
                # Pivot ij is difficult, pivot jk is easy
                self.threshold_utility = 1
                self.justification = self.DIFFICULT_VS_EASY
            else:
                # Both pivots are difficult -> General case of the offset method
                pij = (1 + phi_ik) / (1 - phi_k)
                pjk = (1 + phi_j) * phi_i**2 / (1 - phi_i)
                p1t = phi_i
                p2t = phi_ij
                self.threshold_utility = (pij / 2 + p1t / 3 + p2t / 6) / (pij / 2 + pjk / 2 + p1t * 2 / 3 + p2t / 3)
                self.justification = self.OFFSET_METHOD
        if isnan(self.threshold_utility):
            self.ballot = INCONCLUSIVE  # pragma: no cover
            raise ValueError('Unable to compute threshold utility')  # pragma: no cover
        elif isclose(self.threshold_utility, 1):
            self.ballot = ballot_one(ranking)
        elif isclose(self.threshold_utility, 0):
            self.ballot = ballot_one_two(ranking)
        else:
            self.ballot = UTILITY_DEPENDENT

    def __repr__(self):
        return '<' + ', '.join([
            'ballot = %s' % self.ballot,
            'threshold_utility = {:.6g}'.format(self.threshold_utility),
            'justification = %s' % self.justification,
            'pivot_tij = %s' % self.pivot_tij.asymptotic,
            'pivot_tjk = %s' % self.pivot_tjk.asymptotic,
            'trio_1t = %s' % self.trio_1t.asymptotic,
            'trio_2t = %s' % self.trio_2t.asymptotic,
            'trio = %s' % self.trio.asymptotic,
        ]) + '>'



from fractions import Fraction
from poisson_approval import ProfileOrdinal
from poisson_approval import StrategyOrdinal


def test():
    """
    >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
    >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'}, profile=profile)
    >>> strategy_response = profile.best_responses_to_strategy(strategy.d_ranking_best_response)
    >>> print(strategy_response)
    <abc: a, bac: ab, cab: c> ==> a
    """
    pass



def _cache(f):
    """Auxiliary decorator used by :meth:`cached_property`.

    Parameters
    ----------
    f : callable
        A method with no argument (except ``self``).

    Returns
    -------
    callable
        The same function, but with a `caching` behavior.

    Notes
    -----
    In practice, the values are stored in an attribute of the object that is a dictionary called ``_cached_properties``.
    For example, consider a method ``foo(self)``.

        * If the value is not computed yet, the decorated method will compute the value, store it in
          ``self._cached_properties['foo']`` and return it.
        * If the value is already computed, the decorated method will get it from ``self._cached_properties['foo']``
          and return it.
    """
    name = f.__name__

    # noinspection PyProtectedMember
    def _f(*args):
        try:
            return args[0]._cached_properties[name]
        except KeyError:
            # Not stored in cache
            value = f(*args)
            args[0]._cached_properties[name] = value
            return value
        except AttributeError:
            # cache does not even exist
            value = f(*args)
            args[0]._cached_properties = {name: value}
            return value
    _f.__doc__ = f.__doc__
    return _f


def cached_property(f):
    """Decorator used in replacement of ``@property`` to put the value in cache automatically.

    Notes
    -----
    The first time the attribute is used, it is computed on-demand and put in cache. Later accesses to the
    attributes will use the cached value.

    Cf. :class:`DeleteCacheMixin` for an example.
    """
    return property(_cache(f))


class DeleteCacheMixin:
    """Mixin used to delete cached properties.

    Notes
    -----
    Cf. decorator :meth:`cached_property`.

    Examples
    --------
        >>> class Example(DeleteCacheMixin):
        ...     @cached_property
        ...     def x(self):
        ...         print('Big computation...')
        ...         return 6 * 7
        >>> a = Example()
        >>> a.x
        Big computation...
        42
        >>> a.x
        42
        >>> a.delete_cache()
        >>> a.x
        Big computation...
        42
    """

    # noinspection PyAttributeOutsideInit
    def delete_cache(self) -> None:
        self._cached_properties = dict()



class SetPrintingInOrder(set):
    """A set that prints in order.

    Examples
    --------
        >>> s = SetPrintingInOrder({'b', 'a', 'c'})
        >>> print(s)
        {a, b, c}
        >>> print(repr(s))
        {'a', 'b', 'c'}
    """

    def __repr__(self):
        return "{" + ", ".join([
             "%r" % element for element in sorted(self)
        ]) + "}"

    def __str__(self):
        return "{" + ", ".join([
             "%s" % element for element in sorted(self)
        ]) + "}"

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')



import numpy as np
from poisson_approval.constants.constants import *
from poisson_approval.utils.SetPrintingInOrder import SetPrintingInOrder
from poisson_approval.containers.Winners import Winners
from poisson_approval.utils.UtilCache import cached_property
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold


# noinspection PyUnresolvedReferences
class Profile():
    """A profile of preference (abstract class)."""

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    @cached_property
    def d_ranking_share(self):
        """dict : Shares of rankings.
            E.g. ``'abc': 0.3`` means that a ratio 0.3 of the voters have ranking ``'abc'``.
        """
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError

    @cached_property
    def standardized_version(self):
        """Profile : Standardized version of the profile (makes it unique, up to permutations of the candidates)."""
        raise NotImplementedError

    @cached_property
    def is_standardized(self):
        """bool : Whether the profile is standardized. Cf. :meth:`standardized_version`."""
        return self == self.standardized_version

    # Condorcet stuff

    @cached_property
    def weighted_maj_graph(self):
        """np.ndarray : Weighted majority graph."""
        ab = self.abc + self.acb + self.cab - self.bac - self.bca - self.cba
        ac = self.abc + self.acb + self.bac - self.bca - self.cab - self.cba
        bc = self.abc + self.bac + self.bca - self.acb - self.cab - self.cba
        return np.array([[0, ab, ac], [-ab, 0, bc], [-ac, -bc, 0]])

    @cached_property
    def condorcet_winners(self):
        """Winners : Condorcet winner(s)."""
        m = self.weighted_maj_graph
        min_score = {'a': min(m[0, 1], m[0, 2]), 'b': min(m[1, 0], m[1, 2]), 'c': min(m[2, 0], m[2, 1])}
        return Winners({candidate for candidate in ['a', 'b', 'c'] if min_score[candidate] > - 10**(-8)})

    @cached_property
    def is_profile_condorcet(self):
        """float : Whether the profile is Condorcet. 1. means there is a strict Condorcet winner, 0.5 means there are
        one or more weak Condorcet winner(s), 0. means there is no Condorcet winner.
        """
        m = self.weighted_maj_graph
        min_score = [min(m[0, 1], m[0, 2]), min(m[1, 0], m[1, 2]), min(m[2, 0], m[2, 1])]
        maximin = max(min_score)
        if maximin > 10**(-8):
            return 1.
        elif maximin > - 10**(-8):
            return .5
        else:
            return 0.

    @cached_property
    def has_majority_favorite(self):
        """bool : Whether there is a `majority favorite` (a candidate ranked first by strictly more than half of the
        voters).
        """
        return self.abc + self.acb > 0.5 or self.bac + self.bca > 0.5 or self.cab + self.cba > 0.5

    @cached_property
    def has_majority_ranking(self):
        """bool : Whether there is a majority ranking (a ranking shared by strictly more than half of the voters)."""
        return max(self.d_ranking_share.values()) > 0.5

    # Single-peakedness
    @cached_property
    def is_single_peaked(self):
        """bool : Whether the profile is single-peaked."""
        return ((self.abc == 0 and self.bac == 0)
                or (self.acb == 0 and self.cab == 0)
                or (self.bca == 0 and self.cba == 0))

    # Has full support
    @cached_property
    def support_in_rankings(self):
        """:class:`SetPrintingInOrder` of str : Support of the profile (in terms of rankings)."""
        return SetPrintingInOrder({key for key, val in self.d_ranking_share.items() if val > 0})

    @cached_property
    def is_generic_in_rankings(self):
        """bool : Whether the profile is generic in rankings (contains all rankings)."""
        return 0 not in self.d_ranking_share.values()

    # Tau and strategy-related stuff

    def tau(self, strategy):
        """Tau-vector associated to this profile and a given strategy.

        Parameters
        ----------
        strategy : Strategy
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.
        """
        raise NotImplementedError

    # noinspection NonAsciiCharacters
    def τ(self, strategy):
        """Tau-vector (alternate notation).

        Parameters
        ----------
        strategy : Strategy
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.
        """
        return self.tau(strategy)

    def is_equilibrium(self, strategy):
        """Whether a strategy is an equilibrium in this profile.

        Parameters
        ----------
        strategy : Strategy
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        EquilibriumStatus
            Whether `strategy` is an equilibrium in this profile.
        """
        raise NotImplementedError

    def best_responses_to_strategy(self, d_ranking_best_response):
        """Convert best responses to a :class:`StrategyThreshold`.

        Parameters
        ----------
        d_ranking_best_response : dict
            Key: ranking. Value: :class:`BestResponse`.

        Returns
        -------
        StrategyThreshold
            The conversion of the best responses into a strategy. Only the rankings present in this profile are
            mentioned in the strategy.
        """
        return StrategyThreshold({
            ranking: best_response.threshold_utility
            for ranking, best_response in d_ranking_best_response.items()
            if self.d_ranking_share[ranking] > 0
        }, profile=self)


def make_property_ranking_share(ranking, doc):
    def _f(self):
        return self.d_ranking_share[ranking]
    _f.__doc__ = doc
    return property(_f)


for my_ranking in RANKINGS:
    setattr(Profile, my_ranking, make_property_ranking_share(my_ranking, 'Number : Share of voters with this ranking.'))



from poisson_approval import NiceStatsProfileOrdinal, initialize_random_seeds, ProfileOrdinal


def test():
    # Test when rejection is indeed necessary.
    initialize_random_seeds()
    nice_stats = NiceStatsProfileOrdinal(
        tests_profile=[
            (lambda profile: any([strategy for strategy in profile.analyzed_strategies.equilibria
                                  if strategy.profile.condorcet_winners == strategy.winners]),
             'There exists a true equilibrium electing the CW'),
        ],
        conditional_on=lambda profile: profile.has_majority_ranking)
    nice_stats.run(n_samples=2)
    profile = nice_stats.find_example(0, False)
    assert isinstance(profile, ProfileOrdinal)



import numpy as np
from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event
from poisson_approval.events.EventPivotWeak import EventPivotWeak


class EventPivotTjk(Event):
    """A `personalized pivot` of type ``Tjk``.

    Notes
    -----
    We consider the personalized pivot for the two least-liked candidates of a voter ``zyx``, i.e. situations
    where, if the voter add a reasonable ballot (``z`` or ``yz``), it become a strict pivot for ``xy``. In other words,
    situations where ``S_x = S_y > S_z + 1`` or ``S_x = S_y + 1 > S_z + 1``.

    For parameters and attributes, cf. :class:`Event`.

    Examples
    --------
        >>> from fractions import Fraction
        >>> EventPivotTjk(candidate_x='a', candidate_y='b', candidate_z='c',
        ...               tau_a=Fraction(1, 10), tau_ab=Fraction(6, 10), tau_c=Fraction(3, 10))
        <asymptotic = exp(- 0.1 n + log n - 2.30259 + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    """

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        pivot_weak = EventPivotWeak(candidate_x='x', candidate_y='y', candidate_z='z',
                                    tau_x=tau_x, tau_y=tau_y, tau_z=tau_z, tau_xy=tau_xy, tau_xz=tau_xz, tau_yz=tau_yz)
        self._phi_x = pivot_weak.phi['x']
        self._phi_y = pivot_weak.phi['y']
        self._phi_z = pivot_weak.phi['z']
        self._phi_xy = pivot_weak.phi['xy']
        self._phi_xz = pivot_weak.phi['xz']
        self._phi_yz = pivot_weak.phi['yz']
        if tau_xy == 0 and (tau_x == 0 or tau_y == 0):
            # Flower diagram: holes `at the bottom`, i.e. around ``xy``
            self.asymptotic = Asymptotic(mu=-np.inf, nu=-np.inf, xi=-np.inf)
        elif tau_x == 0 and tau_xz == 0:
            # Flower diagram: consecutive holes on the ``x`` side
            self.asymptotic = (Asymptotic.poisson_value(tau_yz, 0) * Asymptotic.poisson_value(tau_y, 0)
                               * Asymptotic.poisson_gt_one_more(tau_xy, tau_z))
        elif tau_y == 0 and tau_yz == 0:
            # Flower diagram: consecutive holes on the ``y`` side
            self.asymptotic = (
                Asymptotic.poisson_value(tau_x, 0)
                * (Asymptotic.poisson_value(tau_xz, 0) + Asymptotic.poisson_value(tau_xz, 1))
                * Asymptotic.poisson_gt_one_more(tau_xy, tau_z)
            ) + (
                Asymptotic.poisson_value(tau_x, 1)
                * Asymptotic.poisson_value(tau_xz, 0)
                * Asymptotic.poisson_gt(tau_xy, tau_z)
            )
        elif tau_z == 0 and (tau_xz == 0 or tau_yz == 0):
            # Flower diagram: holes `at the top`, i.e. around ``z``
            self.asymptotic = (Asymptotic.poisson_eq(tau_x + tau_xz, tau_y + tau_yz)
                               + Asymptotic.poisson_one_more(tau_x + tau_xz, tau_y + tau_yz))
        else:
            _phi_z_tilde = self._phi_z if tau_z != 0 else self._phi_xz * self._phi_yz
            _phi_y_tilde = self._phi_y if tau_y != 0 else self._phi_xy * self._phi_yz
            self.asymptotic = pivot_weak.asymptotic * (_phi_z_tilde**2 * (1 + _phi_y_tilde))
        if np.isinf(float(self.asymptotic.mu)):
            self._phi_x = np.nan
            self._phi_y = np.nan
            self._phi_z = np.nan
            self._phi_xy = np.nan
            self._phi_xz = np.nan
            self._phi_yz = np.nan



import random
from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import initialize_random_seeds, ballot_one, ballot_one_two
from poisson_approval.strategies.StrategyOrdinal import StrategyOrdinal


class GeneratorStrategyOrdinalUniform:
    """A generator of ordinal strategies (:class:`StrategyOrdinal`) following the uniform distribution.

    Notes
    -----
    Each ballot is chosen at random: either the top candidate or the two top candidates.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> generator = GeneratorStrategyOrdinalUniform()
        >>> strategy = generator()
        >>> print(strategy)
        <abc: ab, acb: ac, bac: b, bca: bc, cab: ac, cba: bc>
    """

    def __call__(self):
        """
        Returns
        -------
        StrategyOrdinal
            A strategy.
        """
        return StrategyOrdinal(d_ranking_ballot={
            ranking: random.choice([ballot_one(ranking), ballot_one_two(ranking)])
            for ranking in RANKINGS})



from math import isclose
from fractions import Fraction
from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import ballot_one, ballot_one_two, barycenter
from poisson_approval.profiles.Profile import Profile
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.utils.UtilCache import cached_property
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold


# noinspection PyUnresolvedReferences
class ProfileCardinal(Profile):
    """A cardinal profile of preference (abstract class).

    Parameters
    ----------
    ratio_sincere : Number
        The ratio of sincere voters, in the interval [0, 1]. This is used for :meth:`tau`.
    """

    def __init__(self, ratio_sincere=0):
        super().__init__()
        self.ratio_sincere = ratio_sincere

    def have_ranking_with_utility_above_u(self, ranking, u):
        """Share of voters who have a given ranking and strictly above a given utility for their middle candidate.

        Parameters
        ----------
        ranking : str
            A ranking, e.g. ``'abc'``.
        u : Number
            A utility between 0 and 1 (included).

        Returns
        -------
        Number
            The share of voters who have ranking `ranking` and a utility for their middle candidate strictly greater
            than `u`.
        """
        raise NotImplementedError

    def have_ranking_with_utility_u(self, ranking, u):
        """Share of voters who have a given ranking and a given utility for their middle candidate.

        Parameters
        ----------
        ranking : str
            A ranking, e.g. ``'abc'``.
        u : Number
            A utility between 0 and 1 (included).

        Returns
        -------
        Number
            The share of voters who have ranking `ranking` and a utility for their middle candidate equal to `u`.
        """
        raise NotImplementedError

    def have_ranking_with_utility_below_u(self, ranking, u):
        """Share of voters who have a given ranking and strictly below a given utility for their middle candidate.

        Parameters
        ----------
        ranking : str
            A ranking, e.g. ``'abc'``.
        u : Number
            A utility between 0 and 1 (included).

        Returns
        -------
        Number
            The share of voters who have ranking `ranking` and a utility for their middle candidate strictly lower
            than `u`.
        """
        raise NotImplementedError

    @cached_property
    def d_ranking_share(self):
        return DictPrintingInOrderIgnoringZeros({
            ranking: self.have_ranking_with_utility_above_u(ranking, 0) + self.have_ranking_with_utility_u(ranking, 0)
            for ranking in RANKINGS
        })

    def __eq__(self, other):
        raise NotImplementedError

    @cached_property
    def standardized_version(self):
        raise NotImplementedError

    # Tau and strategy-related stuff

    def tau(self, strategy):
        """Tau-vector associated to a strategy, with partial sincere voting.

        Parameters
        ----------
        strategy : an argument accepted by :meth:`tau_strategic`

        Returns
        -------
        TauVector
            A share :attr:`ratio_sincere` of the voters vote sincerely (in the sense of :attr:`tau_sincere`) and the
            rest of them vote strategically (in the sense of :meth:`tau_strategic`). In other words, this tau-vector
            is the barycenter of ``tau_sincere`` and ``tau_strategic(strategy)``, with respective weights
            ``self.ratio_sincere`` and ``1 - self.ratio_sincere``.
        """
        tau_sincere = self.tau_sincere
        tau_strategic = self.tau_strategic(strategy)
        t = {ballot: barycenter(a=tau_strategic.d_ballot_share[ballot],
                                b=tau_sincere.d_ballot_share[ballot],
                                ratio_b=self.ratio_sincere)
             for ballot in BALLOTS_WITHOUT_INVERSIONS}
        return TauVector(t)

    @cached_property
    def tau_sincere(self):
        """Tau-vector associated to sincere voting.

        Returns
        -------
        TauVector
            All voters approve of their top candidate. Voters approve of their middle candidate if and only if
            their utility for her is strictly greater than 0.5.
        """
        t = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for ranking in RANKINGS:
            share_vote_one_two = self.have_ranking_with_utility_above_u(ranking, Fraction(1, 2))
            share_vote_one = self.d_ranking_share[ranking] - share_vote_one_two
            t[ballot_one(ranking)] += share_vote_one
            t[ballot_one_two(ranking)] += share_vote_one_two
        return TauVector(t)

    def tau_strategic(self, strategy):
        """Tau-vector associated to a strategy (fully strategic voting).

        Parameters
        ----------
        strategy : StrategyThreshold
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.
        """
        t = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for ranking, threshold in strategy.d_ranking_threshold.items():
            if self.d_ranking_share[ranking] == 0:
                continue
            t[ballot_one(ranking)] += (self.have_ranking_with_utility_u(ranking, u=threshold)
                                       + self.have_ranking_with_utility_below_u(ranking, u=threshold))
            t[ballot_one_two(ranking)] += self.have_ranking_with_utility_above_u(ranking, u=threshold)
        return TauVector(t)

    def is_equilibrium(self, strategy):
        """Whether a strategy is an equilibrium.

        Parameters
        ----------
        strategy : StrategyThreshold
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        EquilibriumStatus
            Whether `strategy` is an equilibrium in this profile. This is based on the assumption that:

            * A proportion :attr:`ratio_sincere` of voters cast their ballot sincerely (in the sense of
              :attr:`tau_sincere`),
            * And the rest of the voters use `strategy`.
        """
        this_tau = self.tau(strategy)
        d_ranking_best_response = this_tau.d_ranking_best_response
        d_ranking_threshold = dict()
        for ranking, share in self.d_ranking_share.items():
            if share == 0:
                continue
            best_response = d_ranking_best_response[ranking]
            if best_response.ballot == INCONCLUSIVE:  # pragma: no cover
                return EquilibriumStatus.INCONCLUSIVE
            d_ranking_threshold[ranking] = best_response.threshold_utility
        tau_response = self.tau(StrategyThreshold(d_ranking_threshold))
        if tau_response.isclose(this_tau):
            return EquilibriumStatus.EQUILIBRIUM
        else:
            return EquilibriumStatus.NOT_EQUILIBRIUM

    def iterated_voting_strategies(self, strategy_ini, n_max_episodes, update_ratio=1, verbose=False):
        """Seek for convergence by iterated voting (strategy update).

        Parameters
        ----------
        strategy_ini : StrategyThreshold
            Initial strategy.
        n_max_episodes : int
            Maximal number of iterations.
        update_ratio : Number
            The speed at which the utility threshold of the strategy being used moves toward the utility threshold of
            the best response. For example, for voters `abc`, if the current threshold of their strategy is `t` and the
            threshold of their best response is `u`, then the threshold of their updated strategy will be
            ``(1 - update_ratio) * t + update_ratio * u``.
        verbose : bool
            If True, print all intermediate strategies.

        Returns
        -------
        dict
            * Key ``cycle``: a list of :class:`TauVector`. The limit cycle of tau-vectors. If its length is 1, the
              process converges to this tau-vector. If its length is greater than 1, the process reaches a periodical
              orbit between these tau-vectors. If its length is 0, by convention, it means that the process does not
              converge and does not reach a periodical orbit.
            * Key ``responses``: a list of :class:`StrategyThreshold`. Its length is the same as ``cycle``. Each
              element represent the strategy that is the best response to the corresponding tau-vector. For example,
              if ``update_ratio == 1``, it is the strategy that all strategic voters use at the following step of the
              cycle.
        """
        strategy = StrategyThreshold({
            ranking: threshold for ranking, threshold in strategy_ini.d_ranking_threshold.items()
            if self.d_ranking_share[ranking] > 0
        }, profile=self)
        tau = strategy.tau
        taus = [tau]
        if verbose:
            print(-1)
            print(strategy)
        for i in range(n_max_episodes):
            strategy = StrategyThreshold(
                {ranking: _my_round(barycenter(a=strategy.d_ranking_threshold[ranking],
                                               b=strategy.d_ranking_best_response[ranking].threshold_utility,
                                               ratio_b=update_ratio))
                 for ranking in RANKINGS if self.d_ranking_share[ranking] > 0},
                profile=self)
            tau = strategy.tau
            if verbose:
                print(i)
                print(strategy)
            if tau in taus:
                # If there is an exact cycle, it is useless to continue looping.
                taus.append(tau)
                break
            else:
                taus.append(tau)
        try:
            end = len(taus) - 1
            begin = next(begin
                         for begin in range(end - 1, -1, -1)
                         if taus[begin].isclose(taus[end]))
            cycle_taus = taus[begin:end]
        except StopIteration:
            cycle_taus = []
        responses = [self.best_responses_to_strategy(tau.d_ranking_best_response) for tau in cycle_taus]
        return {'cycle_taus': cycle_taus, 'responses': responses}

    def iterated_voting_taus(self, strategy_ini, n_max_episodes, update_ratio=1, verbose=False):
        """Seek for convergence by iterated voting (tau update).

        Parameters
        ----------
        strategy_ini : an argument accepted by :meth:`tau`, i.e. by :meth:`tau_strategic`
            The initial strategy.
        n_max_episodes : int
            Maximal number of iterations.
        update_ratio : Number
            The ratio of voters that update their ballot. At each iteration, this ratio of voters update their response
            (they continue voting sincerely if they are sincere voters, they update their strategical ballot otherwise).
            In other words, the updated tau-vector is a barycenter of the former tau-vector with weight
            ``1 - update_ratio`` and the response tau-vector (as computed by :meth:``tau``) with weight
            ``update_ratio``.
        verbose : bool
            If True, print all intermediate tau-vectors.

        Returns
        -------
        dict
            * Key ``cycle``: a list of :class:`TauVector`. The limit cycle of tau-vectors. If its length is 1, the
              process converges to this tau-vector. If its length is greater than 1, the process reaches a periodical
              orbit between these tau-vectors. If its length is 0, by convention, it means that the process does not
              converge and does not reach a periodical orbit.
            * Key ``responses``: a list of :class:`StrategyThreshold`. Its length is the same as ``cycle``. Each
              element represent the strategy that is the best response to the corresponding tau-vector. For example,
              if ``update_ratio == 1``, it is the strategy that all strategic voters use at the following step of the
              cycle.
        """
        tau = self.tau(strategy_ini)
        taus = [tau]
        if verbose:
            print(-1)
            print(tau)
        for i in range(n_max_episodes):
            strategy = self.best_responses_to_strategy(tau.d_ranking_best_response)
            tau_full_response = strategy.tau
            tau = TauVector({
                ballot: _my_round(barycenter(a=tau.d_ballot_share[ballot],
                                             b=tau_full_response.d_ballot_share[ballot],
                                             ratio_b=update_ratio))
                for ballot in BALLOTS_WITHOUT_INVERSIONS
            }, normalization_warning=False)
            if verbose:
                print(i)
                print(tau)
            if tau in taus:
                # If there is an exact cycle, it is useless to continue looping.
                taus.append(tau)
                break
            else:
                taus.append(tau)
        try:
            end = len(taus) - 1
            begin = next(begin
                         for begin in range(end - 1, -1, -1)
                         if taus[begin].isclose(taus[end]))
            cycle_taus = taus[begin:end]
        except StopIteration:
            cycle_taus = []
        responses = [self.best_responses_to_strategy(tau.d_ranking_best_response) for tau in cycle_taus]
        return {'cycle_taus': cycle_taus, 'responses': responses}


def _my_round(x):
    """Round to 0 or 1 if the number is close enough.

    Parameters
    ----------
    x : Number

    Returns
    -------
    Number
        Generally, return `x`. If `x` is close enough to 0 or 1, then return 0 or 1 respectively.

    Examples
    --------

        >>> _my_round(1E-10)
        0
        >>> _my_round(0.999999999999999999)
        1
        >>> _my_round(0.123456789)
        0.123456789
    """
    if isclose(x, 1):
        return 1
    if isclose(x, 0, abs_tol=1E-9):
        return 0
    return x



class EquilibriumStatus:
    """
    The equilibrium status.

    The possible values are:

    * ``EquilibriumStatus.EQUILIBRIUM``,
    * ``EquilibriumStatus.UTILITY_DEPENDENT``,
    * ``EquilibriumStatus.INCONCLUSIVE``,
    * ``EquilibriumStatus.NOT_EQUILIBRIUM``.

    Examples
    --------
        >>> min(EquilibriumStatus.EQUILIBRIUM, EquilibriumStatus.INCONCLUSIVE)
        EquilibriumStatus.INCONCLUSIVE
    """

    EQUILIBRIUM = None
    UTILITY_DEPENDENT = None
    INCONCLUSIVE = None
    NOT_EQUILIBRIUM = None

    def __init__(self, n, s, r):
        self.n = n
        self.s = s
        self.r = r

    def __repr__(self):
        """
        Examples
        --------
            >>> repr(EquilibriumStatus.EQUILIBRIUM)
            'EquilibriumStatus.EQUILIBRIUM'
        """
        return self.r

    def __str__(self):
        """
        Examples
        --------
            >>> str(EquilibriumStatus.EQUILIBRIUM)
            'equilibrium'
        """
        return self.s

    def __eq__(self, other):
        return self.n == other.n

    def __lt__(self, other):
        return self.n < other.n


EquilibriumStatus.EQUILIBRIUM = EquilibriumStatus(3, 'equilibrium', 'EquilibriumStatus.EQUILIBRIUM')
EquilibriumStatus.UTILITY_DEPENDENT = EquilibriumStatus(2, 'utility-dependent', 'EquilibriumStatus.UTILITY_DEPENDENT')
EquilibriumStatus.INCONCLUSIVE = EquilibriumStatus(1, 'inconclusive', 'EquilibriumStatus.INCONCLUSIVE')
EquilibriumStatus.NOT_EQUILIBRIUM = EquilibriumStatus(0, 'not equilibrium', 'EquilibriumStatus.NOT_EQUILIBRIUM')



import numpy as np
from fractions import Fraction
from poisson_approval import ProfileTwelve, StrategyTwelve, StrategyThreshold, EquilibriumStatus


def test():
    """
    Iterated voting with verbose option:

        >>> from fractions import Fraction
        >>> profile = ProfileTwelve(d_type_share={'a_bc': Fraction(1, 2), 'ab_c': Fraction(1, 2)})
        >>> strategy = StrategyThreshold(d_ranking_threshold={'abc': 0})
        >>> result = profile.iterated_voting_strategies(strategy_ini=strategy, n_max_episodes=100,
        ...                                             update_ratio=1, verbose=True)
        -1
        <abc: ab> ==> a, b
        0
        <abc: a> ==> a
        1
        <abc: a> ==> a
        >>> result = profile.iterated_voting_taus(strategy_ini=strategy, n_max_episodes=100,
        ...                                       update_ratio=1, verbose=True)
        -1
        <ab: 1> ==> a, b
        0
        <a: 1> ==> a
        1
        <a: 1> ==> a
    """

    # Check that normalization works as expected
    profile = ProfileTwelve(d_type_share={'a_bc': 1, 'ab_c': 1})
    assert profile.a_bc == 0.5

    # Check a strategy that is not an equilibrium
    profile = ProfileTwelve(d_type_share={'a_bc': 1, 'ab_c': 1})
    strategy = StrategyTwelve(d_ranking_ballot={'abc': 'ab'})
    assert profile.is_equilibrium(strategy) == EquilibriumStatus.NOT_EQUILIBRIUM

    # Case without convergence: iterated_voting_strategies
    profile = ProfileTwelve(d_type_share={'a_bc': 1, 'ab_c': 1})
    strategy = StrategyThreshold(d_ranking_threshold={'abc': 0})
    result = profile.iterated_voting_strategies(strategy_ini=strategy, n_max_episodes=1, update_ratio=1)
    assert result['cycle_taus'] == []

    # Case without convergence: iterated_voting_taus
    profile = ProfileTwelve(d_type_share={'a_bc': 1, 'ab_c': 1})
    strategy = StrategyThreshold(d_ranking_threshold={'abc': 0})
    result = profile.iterated_voting_taus(strategy_ini=strategy, n_max_episodes=1, update_ratio=1)
    assert result['cycle_taus'] == []



import numpy as np
from fractions import Fraction
from poisson_approval import ProfileOrdinal


def test():
    # Check that normalization works as expected
    profile = ProfileOrdinal(d_ranking_share={'abc': 1, 'acb': 1})
    assert profile.abc == 0.5



import warnings
import numpy as np
from math import isclose
from poisson_approval.constants.constants import *
from poisson_approval.profiles.Profile import Profile
from poisson_approval.utils.UtilMasks import masks_area, masks_distribution, winners_distribution
from poisson_approval.strategies.StrategyOrdinal import StrategyOrdinal
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.containers.AnalyzedStrategies import AnalyzedStrategies
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.utils.UtilCache import cached_property


# noinspection PyUnresolvedReferences
class ProfileOrdinal(Profile):
    """An ordinal profile of preference.

    Parameters
    ----------
    d_ranking_share : dict
        E.g. ``{'abc': 0.4, 'cab': 0.6}``. ``d_ranking_share['abc']`` is the probability that a voter prefers
        candidate ``a``, then candidate ``b``, then candidate ``c``.
    normalization_warning : bool
        Whether a warning should be issued if the input distribution is not normalized.
    well_informed_voters : bool.
        If True (default), it is the usual model. If False, voters "see" only the candidates' expected scores and
        believe that the scores follow independent Poisson distributions.

    Notes
    -----
    If the input distribution `d_ranking_share` is not normalized, the profile will be normalized anyway and a
    warning is issued (unless `normalization_warning` is False).

    Examples
    --------
        >>> from fractions import Fraction
        >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
        >>> profile
        ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(3, 5), 'cab': Fraction(3, 10)})
        >>> print(profile)
        <abc: 1/10, bac: 3/5, cab: 3/10> (Condorcet winner: b)
        >>> profile.abc
        Fraction(1, 10)
        >>> profile.d_ranking_share['abc']  # Alternate syntax for profile.abc
        Fraction(1, 10)
        >>> profile.weighted_maj_graph
        array([[0, Fraction(-1, 5), Fraction(2, 5)],
               [Fraction(1, 5), 0, Fraction(2, 5)],
               [Fraction(-2, 5), Fraction(-2, 5), 0]], dtype=object)
        >>> profile.condorcet_winners
        Winners({'b'})
        >>> profile.is_profile_condorcet
        1.0
        >>> profile.has_majority_favorite  # Is one candidate 'top' in a majority of ballots?
        True
        >>> profile.has_majority_ranking  # Does one ranking represent a majority of ballots?
        True
        >>> profile.is_single_peaked  # Is the profile single-peaked?
        True
        >>> profile.support_in_rankings
        {'abc', 'bac', 'cab'}
        >>> profile.is_generic_in_rankings  # Are all rankings there?
        False
    """

    def __init__(self, d_ranking_share, normalization_warning=True, well_informed_voters=True):
        """
            >>> profile = ProfileOrdinal({'non_existing_ranking': 1})
            Traceback (most recent call last):
            ValueError: Unknown key: non_existing_ranking
        """
        super().__init__()
        # Populate the dictionary and check for typos in the input
        self._d_ranking_share = DictPrintingInOrderIgnoringZeros()
        for ranking, share in d_ranking_share.items():
            if ranking in RANKINGS:
                self._d_ranking_share[ranking] = share
            else:
                raise ValueError('Unknown key: ' + ranking)
        for ranking in RANKINGS:
            if ranking not in self._d_ranking_share:
                self._d_ranking_share[ranking] = 0
        # Normalize if necessary
        total = sum(self._d_ranking_share.values())
        if not isclose(total, 1.):
            if normalization_warning:
                warnings.warn("Warning: profile is not normalized, I will normalize it.")
            for ranking in self._d_ranking_share.keys():
                self._d_ranking_share[ranking] = self._d_ranking_share[ranking] / total
        # Well-informed voters?
        self.well_informed_voters = well_informed_voters

    @cached_property
    def d_ranking_share(self):
        return self._d_ranking_share

    def __repr__(self):
        return 'ProfileOrdinal(%r)' % self._d_ranking_share

    def __str__(self):
        result = '<%s>' % str(self._d_ranking_share)[1:-1]
        if self.is_profile_condorcet:
            result += ' (Condorcet winner: %s)' % self.condorcet_winners
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : Object

        Returns
        -------
        bool
            True iff this profile is equal to `other`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile == ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            True
        """
        return isinstance(other, ProfileOrdinal) and self.d_ranking_share == other.d_ranking_share

    @cached_property
    def standardized_version(self):
        """ProfileOrdinal : Standardized version of the profile (makes it unique, up to permutations of the candidates).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.standardized_version
            ProfileOrdinal({'abc': Fraction(3, 5), 'bac': Fraction(1, 10), 'cba': Fraction(3, 10)})
            >>> profile.is_standardized
            False
        """
        def translate(s, permute):
            return s.replace('a', permute[0]).replace('b', permute[1]).replace('c', permute[2])

        best_d = {}
        best_signature = []
        for perm in XYZ_PERMUTATIONS:
            d_test = {translate(ranking, perm): share for ranking, share in self.d_ranking_share.items()}
            signature_test = [d_test[ranking] for ranking in XYZ_RANKINGS]
            if signature_test > best_signature:
                best_signature = signature_test
                best_d = d_test
        return ProfileOrdinal({ranking: best_d[xyz_ranking] for ranking, xyz_ranking in zip(RANKINGS, XYZ_RANKINGS)})

    # Has full support
    @cached_property
    def support(self):
        """:class:`SetPrintingInOrder` of str : Alternate notation for :meth:`support_in_rankings`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.support
            {'abc', 'bac', 'cab'}
        """
        return self.support_in_rankings

    @cached_property
    def is_generic(self):
        """bool : Alternate notation for :meth:`is_generic_in_rankings`

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.is_generic
            False
        """
        return self.is_generic_in_rankings

    # Tau and strategy-related stuff

    def tau(self, strategy):
        """Tau-vector associated to a strategy.

        Parameters
        ----------
        strategy : StrategyOrdinal
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
            >>> tau = profile.tau(strategy)
            >>> print(tau)
            <a: 1/10, ab: 3/5, c: 3/10> ==> a
            >>> tau = profile.τ(strategy)  # Alternate notation
            >>> print(tau)
            <a: 1/10, ab: 3/5, c: 3/10> ==> a

        In the case of badly informed voters, we do as if the tau-vector were the vector of scores (up to
        a renormalization):

            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...                          well_informed_voters=False)
            >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
            >>> tau = profile.tau(strategy)
            >>> print(tau)
            <a: 7/16, b: 3/8, c: 3/16> ==> a
        """
        t = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for ranking, ballot in strategy.d_ranking_ballot.items():
            if self.d_ranking_share[ranking] > 0:
                t[ballot] += self.d_ranking_share[ranking]
        if self.well_informed_voters:
            return TauVector(t)
        else:
            return TauVector(
                {'a': t['a'] + t['ab'] + t['ac'], 'b': t['b'] + t['ab'] + t['bc'], 'c': t['c'] + t['ac'] + t['bc']},
                normalization_warning=False)

    def is_equilibrium(self, strategy):
        """Whether a strategy is an equilibrium.

        Parameters
        ----------
        strategy : StrategyOrdinal
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        EquilibriumStatus
            Whether `strategy` is an equilibrium in this profile.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
            >>> profile.is_equilibrium(strategy)
            EquilibriumStatus.EQUILIBRIUM
        """
        d_ranking_best_response = self.tau(strategy).d_ranking_best_response
        status = EquilibriumStatus.EQUILIBRIUM
        for ranking, share in self.d_ranking_share.items():
            if share == 0:
                continue
            best_response = d_ranking_best_response[ranking]
            if best_response.ballot == INCONCLUSIVE:
                status = min(status, EquilibriumStatus.INCONCLUSIVE)  # pragma: no cover
            elif best_response.ballot == UTILITY_DEPENDENT:
                status = min(status, EquilibriumStatus.UTILITY_DEPENDENT)
            elif strategy.d_ranking_ballot[ranking] != best_response.ballot:
                return EquilibriumStatus.NOT_EQUILIBRIUM
        return status

    @cached_property
    def analyzed_strategies(self):
        """AnalyzedStrategies : Analyzed strategies of the profile.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.analyzed_strategies
            Equilibria:
            <abc: a, bac: b, cab: ac> ==> b (FF)
            <abc: a, bac: ab, cab: c> ==> a (D)
            <BLANKLINE>
            Utility-dependent equilibrium:
            <abc: ab, bac: b, cab: c> ==> b (FF)
            <BLANKLINE>
            Non-equilibria:
            <abc: a, bac: b, cab: c> ==> b (FF)
            <abc: a, bac: ab, cab: ac> ==> a (D)
            <abc: ab, bac: b, cab: ac> ==> b (FF)
            <abc: ab, bac: ab, cab: c> ==> a, b (FF)
            <abc: ab, bac: ab, cab: ac> ==> a (D)
            >>> print(profile.analyzed_strategies.equilibria[0])
            <abc: a, bac: b, cab: ac> ==> b
        """
        equilibria = []
        utility_dependent = []
        inconclusive = []
        non_equilibria = []
        strat_abc = ['a', 'ab'] if self.abc > 0 else ['']
        strat_acb = ['a', 'ac'] if self.acb > 0 else ['']
        strat_bac = ['b', 'ab'] if self.bac > 0 else ['']
        strat_bca = ['b', 'bc'] if self.bca > 0 else ['']
        strat_cab = ['c', 'ac'] if self.cab > 0 else ['']
        strat_cba = ['c', 'bc'] if self.cba > 0 else ['']
        for s_abc in strat_abc:
            for s_acb in strat_acb:
                for s_bac in strat_bac:
                    for s_bca in strat_bca:
                        for s_cab in strat_cab:
                            for s_cba in strat_cba:
                                strategy = StrategyOrdinal({'abc': s_abc, 'acb': s_acb, 'bac': s_bac,
                                                         'bca': s_bca, 'cab': s_cab, 'cba': s_cba}, profile=self)
                                status = strategy.is_equilibrium
                                if status == EquilibriumStatus.EQUILIBRIUM:
                                    equilibria.append(strategy)
                                elif status == EquilibriumStatus.UTILITY_DEPENDENT:
                                    utility_dependent.append(strategy)
                                elif status == EquilibriumStatus.INCONCLUSIVE:  # pragma: no cover
                                    inconclusive.append(strategy)
                                    warnings.warn('Met an inconclusive case: \nprofile = %r\nstrategy = %r'
                                                  % (self, strategy))
                                else:
                                    non_equilibria.append(strategy)
        return AnalyzedStrategies(equilibria, utility_dependent, inconclusive, non_equilibria)

    def proba_equilibrium(self, test=None):
        """Probability that an equilibrium exists (depending on the utilities).

        Parameters
        ----------
        test : callable
            A function ``StrategyOrdinal -> bool`` that gives a condition on the strategy. Default: always True.

        Returns
        -------
        float
            The probability that an equilibrium strategy exists, that meets the `test` condition.

        Notes
        -----
        The result is exact (not based on a Monte-Carlo estimation).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.proba_equilibrium()
            1
        """
        if test is None:
            # noinspection PyUnusedLocal
            def test(strategy):
                return True
        if any([test(strategy) for strategy in self.analyzed_strategies.equilibria]):
            return 1
        support = sorted(self.support)
        dim = len(support)
        masks = [
            [(strategy.d_ranking_best_response[ranking].threshold_utility, len(strategy.d_ranking_ballot[ranking]) == 2)
             for ranking in support]
            for strategy in self.analyzed_strategies.utility_dependent if test(strategy)
        ]
        return masks_area(inf=[0] * dim, sup=[1] * dim, masks=masks)

    def distribution_equilibria(self, test=None):
        """Distribution of numbers of equilibria (depending on the utilities).

        Parameters
        ----------
        test : callable
            A function ``StrategyOrdinal -> bool`` that gives a condition on the strategy. Default: always True.

        Returns
        -------
        list
            A list that represents an histogram. The distribution of number of equilibria (meeting the `test` condition).

        Notes
        -----
        The result is exact (not based on a Monte-Carlo estimation).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.distribution_equilibria()
            array([0.        , 0.        , 0.86290531, 0.13709469])
        """
        if test is None:
            # noinspection PyUnusedLocal
            def test(strategy):
                return True
        cover_alls = np.sum([test(strategy) for strategy in self.analyzed_strategies.equilibria], dtype=int)
        support = sorted(self.support)
        dim = len(support)
        masks = [
            [(strategy.d_ranking_best_response[ranking].threshold_utility, len(strategy.d_ranking_ballot[ranking]) == 2)
             for ranking in support]
            for strategy in self.analyzed_strategies.utility_dependent if test(strategy)
        ]
        return masks_distribution(inf=np.zeros(dim), sup=np.ones(dim), masks=masks, cover_alls=cover_alls)

    def distribution_winners(self, test=None):
        """Distribution of the number of equilibrium winners (depending on the utilities).

        Parameters
        ----------
        test : callable
            A function ``StrategyOrdinal -> bool`` that gives a condition on the strategy. Default: always True.

        Returns
        -------
        list
            A list that represents an histogram. The distribution of number of possible equilibrium winner (with
            strategies that meet the `test` condition).

        Notes
        -----
        The result is exact (not based on a Monte-Carlo estimation).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.distribution_winners()
            array([0., 0., 1., 0.])
        """
        if test is None:
            # noinspection PyUnusedLocal
            def test(strategy):
                return True
        cover_alls = set.union(*([set()] + [
            strategy.winners for strategy in self.analyzed_strategies.equilibria
            if test(strategy)
        ]))
        support = sorted(self.support)
        dim = len(support)
        masks_winners = [
            (
                [(strategy.d_ranking_best_response[ranking].threshold_utility,
                  len(strategy.d_ranking_ballot[ranking]) == 2)
                 for ranking in support],
                strategy.winners
            )
            for strategy in self.analyzed_strategies.utility_dependent if test(strategy)
        ]
        return winners_distribution(inf=np.zeros(dim), sup=np.ones(dim), masks_winners=masks_winners,
                                    cover_alls=cover_alls)



import numpy as np
from fractions import Fraction
from poisson_approval import ProfileHistogram, StrategyThreshold, EquilibriumStatus


def test():
    # Check that normalization works as expected
    profile = ProfileHistogram(d_ranking_share={'abc': 1, 'acb': 1},
                               d_ranking_histogram = {'abc': [1, 1], 'acb': [1]})
    assert profile.abc == 0.5
    assert np.all(profile.d_ranking_histogram['abc'] == [0.5, 0.5])

    # Check a strategy that is not an equilibrium
    profile = ProfileHistogram(d_ranking_share={'abc': 0.5, 'acb': 0.5},
                               d_ranking_histogram = {'abc': [1], 'acb': [1]})
    strategy = StrategyThreshold(d_ranking_threshold={'abc': 0, 'acb': 0})
    assert profile.is_equilibrium(strategy) == EquilibriumStatus.NOT_EQUILIBRIUM



from poisson_approval.utils.Util import sort_ballot
from poisson_approval.constants.constants import *
from poisson_approval.strategies.Strategy import Strategy
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros


# noinspection PyUnresolvedReferences
class StrategyTwelve(Strategy):
    """A strategy profile for a 12-type profile.

    Parameters
    ----------
    d_ranking_ballot : dict
        Keys are rankings and values are strategies, e.g. ``'abc': 'ab'``. A strategy can be a valid ballot
        (e.g. ``'a'`` or ``'ab'`` when the ranking is ``'abc'``), ``UTILITY_DEPENDENT`` or ``''`` if the behavior
        of these voters is not specified in the strategy. Cf. :class:`ProfileTwelve`.
    profile : Profile, optional
        The "context" in which the strategy is used.

    Examples
    --------
        >>> strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'})
        >>> strategy
        StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'})
        >>> print(strategy)
        <abc: utility-dependent, bac: b>
        >>> strategy.abc
        'utility-dependent'
        >>> strategy.a_bc
        'a'
        >>> strategy.ab_c
        'ab'
    """

    def __init__(self, d_ranking_ballot, profile=None):
        """
            >>> strategy = StrategyTwelve({'non_existing_ranking': 'utility-dependent'})
            Traceback (most recent call last):
            ValueError: Unknown key: non_existing_ranking
            >>> strategy = StrategyTwelve({'abc': 'non_existing_ballot'})
            Traceback (most recent call last):
            ValueError: Unknown strategy: non_existing_ballot
        """
        # Populate the dictionary and check for typos in the input
        self.d_ranking_ballot = DictPrintingInOrderIgnoringZeros()
        for ranking, ballot in d_ranking_ballot.items():
            if ranking not in RANKINGS:
                raise ValueError('Unknown key: ' + ranking)
            strategy_1 = ranking[0]
            strategy_12 = ranking[:2]
            strategy_21 = ranking[1::-1]
            if ballot not in ['', strategy_1, strategy_12, strategy_21, UTILITY_DEPENDENT]:
                raise ValueError('Unknown strategy: ' + ballot)
            if ballot == UTILITY_DEPENDENT:
                self.d_ranking_ballot[ranking] = ballot
            else:
                self.d_ranking_ballot[ranking] = ''.join(sorted(ballot))
        for ranking in RANKINGS:
            if ranking not in self.d_ranking_ballot:
                self.d_ranking_ballot[ranking] = ''
        # Call parent class
        super().__init__(profile=profile)

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : object

        Returns
        -------
        bool
            True if this strategy is equal to `other`.

        Examples
        --------
            >>> strategy = StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'})
            >>> strategy == StrategyTwelve({'abc': 'utility-dependent', 'bac': 'b'})
            True
        """
        return isinstance(other, StrategyTwelve) and self.d_ranking_ballot == other.d_ranking_ballot

    # Representation

    def __repr__(self):
        return 'StrategyTwelve(%r)' % self.d_ranking_ballot

    def __str__(self):
        result = '<%s>' % str(self.d_ranking_ballot)[1:-1]
        if self.profile is not None:
            result += ' ==> ' + str(self.winners)
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')


def make_property_ranking_ballot(ranking, doc):
    def _f(self):
        return self.d_ranking_ballot[ranking]
    _f.__doc__ = doc
    return property(_f)


for my_ranking in RANKINGS:
    setattr(StrategyTwelve, my_ranking, make_property_ranking_ballot(my_ranking,
                                                                     'str : Ballot of ranking ``%s``.' % my_ranking))


def make_property_ranking_low_u_ballot(ranking, doc):
    def _f(self):
        return ranking[0] if getattr(self, ranking) == UTILITY_DEPENDENT else getattr(self, ranking)
    _f.__doc__ = doc
    return property(_f)


for my_ranking in RANKINGS:
    my_t = my_ranking[0] + '_' + my_ranking[1:]
    setattr(StrategyTwelve, my_t, make_property_ranking_low_u_ballot(my_ranking,
                                                                     'str : Ballot of type ``%s``.' % my_t))


def make_property_ranking_high_u_ballot(ranking, doc):
    def _f(self):
        return sort_ballot(ranking[:2]) if getattr(self, ranking) == UTILITY_DEPENDENT else getattr(self, ranking)
    _f.__doc__ = doc
    return property(_f)


for my_ranking in RANKINGS:
    my_t = my_ranking[:2] + '_' + my_ranking[2]
    setattr(StrategyTwelve, my_t, make_property_ranking_high_u_ballot(my_ranking,
                                                                      'str : Ballot of type ``%s``.' % my_t))
